import json
import re
import os
from .exceptions import NotFoundError, NotAcceptableError
from sgqlc.endpoint.requests import RequestsEndpoint


class HQBase(object):
    _apiprefix = ""

    def __init__(self, client):
        self._hq = client
        self.endpoint = RequestsEndpoint(client.baseurl +'/graphql', session=client.session)

    @property
    def url(self):
        return self._hq.baseurl + self._apiprefix

    def _make_call(self, method, path, filepath = None, parser=None, **kwargs):
        response = self._hq.session.request(method = method, url = path, **kwargs)
        rc = response.status_code
        if rc < 300:
            if method == 'get':
                if 'application/json' in response.headers['Content-Type']:
                    if parser:
                        return parser(response.content)
                    else:
                        return json.loads(response.content)

                elif 'application/zip' in response.headers['Content-Type']:
                    d = response.headers['content-disposition']
                    fname = re.findall(r"filename[ ]*=([^;]+)", d, flags=re.IGNORECASE)

                    if not fname:
                        fname = re.findall(r"filename\*=utf-8''(.+)", d, flags=re.IGNORECASE)
                    fname = fname[0].strip().strip('"')
                    outfile = os.path.join(filepath, fname)

                    with open(outfile, 'wb') as f:
                        for chunk in response.iter_content(chunk_size=1024): 
                            if chunk:
                                f.write(chunk)
                    return outfile
                else:
                    return response.content
            else:
                if response.content:
                    return json.loads(response.content)
                else:
                    return True
        elif rc == 404:
            raise NotFoundError(response.text)
        elif rc == 406:
            raise NotAcceptableError(response.text)
        else:
            response.raise_for_status()



class NotFoundError(Exception):
	def __init__(self, objecttype):
		self.expression = objecttype + ' not found'
		self.message = objecttype + ' not found'

class IncompleteQuestionnaireIdError(Exception):
	def __init__(self):
		self.expression = 'Invalid questionnaire id'
		self.message = 'Both questionnaire guid and version number must be specified'

class UnauthorizedError(Exception):
	def __init__(self):
		self.expression = 'Unauthorized'
		self.message = 'Either username or password were not provided or wrong'

class NotAcceptableError(Exception):
	def __init__(self, message):
		self.expression = 'Not Acceptable'
		self.message = message

class GraphQLError(Exception):
	def __init__(self, message):
		self.expression = 'GraphQL Error'
		self.message = message



from .utils import to_qidentity

class Assignment(object):
    def __init__(self, responsible, quantity, questionnaire_id,
        identifying_data=None, email='', password='', webmode=False,
        audio_recording_enabled=False, comments=''):
        """[summary]
        
        Parameters
        ----------
        responsible : [type]
            [description]
        quantity : [type]
            [description]
        questionnaire_id : [type]
            [description]
        email : str, optional
            [description], by default ''
        password : str, optional
            [description], by default ''
        webmode : bool, optional
            [description], by default False
        audio_recording_enabled : bool, optional
            [description], by default False
        comments : str, optional
            [description], by default ''
        """

        self.responsible = responsible
        self.quantity = quantity
        self.questionnaire_id = questionnaire_id
        self.identifying_data = identifying_data
        self.email = email
        self.password = password
        self.webmode = webmode
        self.audio_recording_enabled = audio_recording_enabled
        self.comments = comments

    def __str__(self):
        return(str(self.__dict__))

    @classmethod
    def from_dict(cls, dict):
        print(dict)
        obj = cls(
            responsible = dict['ResponsibleName'],
            quantity = dict['Quantity'], 
            questionnaire_id = dict['QuestionnaireId'],
            identifying_data=dict['IdentifyingData'] if 'IdentifyingData' in dict else [],
            email = dict['Email'],
            password = dict['Password'],
            webmode = dict['WebMode']
        )
        setattr(obj, 'id', dict['Id'])
        setattr(obj, 'responsible_id', dict['ResponsibleId'])
        setattr(obj, 'interviews_count', dict['InterviewsCount'])
        setattr(obj, 'archived', dict['Archived'])
        setattr(obj, 'created_utc', dict['CreatedAtUtc'])
        setattr(obj, 'updated_utc', dict['UpdatedAtUtc'])
        if 'IsAudioRecordingEnabled' in dict:
            setattr(obj, 'audio_recording_enabled', dict['IsAudioRecordingEnabled'])
        return obj

    def to_json(self):
        return {
                "Responsible": self.responsible,
                "Quantity": self.quantity,
                "QuestionnaireId": self.questionnaire_id,
                "Email": self.email,
                "Password": self.password,
                "WebMode": self.webmode,
                "IsAudioRecordingEnabled": self.audio_recording_enabled,
                "Comments": self.comments
            }

class QuestionnaireListItem(object):
    def __init__(self, dict):
        self.questionnaire_identity = dict['QuestionnaireIdentity']
        self.questionnaire_id = dict['QuestionnaireId']
        self.version = dict['Version']
        self.title = dict['Title']
        self.variable = dict['Variable']
        self.last_entry_date = dict['LastEntryDate']

    def __str__(self):
        return(str(self.__dict__))

    @classmethod
    def from_dict(cls, dict):
        return cls(dict)

class ExportJob(object):
    def __init__(self,
        questionnaire_identity,
        export_type='Tabular',
        interview_status='All',
        from_date=None,
        to_date=None,
        access_token=None,
        storage_type=None):
        """[summary]
        
        Parameters
        ----------
        questionnaire_identity : [type]
            [description]
        export_type : str, optional
            [description], by default 'Tabular'
        interview_status : str, optional
            [description], by default 'All'
        from_date : [type], optional
            [description], by default None
        to_date : [type], optional
            [description], by default None
        access_token : [type], optional
            [description], by default None
        storage_type : [type], optional
            [description], by default None
        """

        if type(questionnaire_identity) is tuple:
            (questionnaire_id, questionnaire_version) = questionnaire_identity 
            questionnaire_identity = to_qidentity(questionnaire_id, questionnaire_version)

        self.questionnaire_identity = questionnaire_identity
        self.export_type = export_type
        self.interview_status = interview_status
        self.from_date = from_date
        self.to_date = to_date
        self.access_token = access_token
        self.storage_type = storage_type

    def __str__(self):
        return(str(self.__dict__))

    @classmethod
    def from_dict(cls, dict):
        obj = cls(
            export_type = dict['ExportType'],
            questionnaire_identity = dict['QuestionnaireId'],
            interview_status = dict['InterviewStatus'],
            from_date = dict['From'] if 'From' in dict else None,
            to_date = dict['To'] if 'To' in dict else None,
            access_token = dict['AccessToken'] if 'AccessToken' in dict else None,
            storage_type = dict['StorageType'] if 'StorageType' in dict else None
        )
        setattr(obj, 'job_id', dict['JobId'])
        setattr(obj, 'export_status', dict['ExportStatus'])
        setattr(obj, 'start_date', dict['StartDate'])
        setattr(obj, 'complete_date', dict['CompleteDate'])
        setattr(obj, 'progress', dict['Progress'])
        if 'ETA' in dict:
            setattr(obj, 'eta', dict['ETA'])
        if 'Links' in dict:
            if 'Cancel' in dict['Links']:
                setattr(obj, 'cancel_link', dict['Links']['Cancel'])
            if 'Download' in dict['Links']:
                setattr(obj, 'download_link', dict['Links']['Download'])
        setattr(obj, 'has_export_file', dict['HasExportFile'])
        return obj

    def to_json(self):
        ret = {
                "ExportType": self.export_type,
                "QuestionnaireId": self.questionnaire_identity,
                "InterviewStatus": self.interview_status,
                "From": self.from_date,
                "To": self.to_date,
                "AccessToken": self.access_token,
                "StorageType": self.storage_type,
            }
        return {k: v for k, v in ret.items() if v is not None}


QUESTION_TYPES = {
	0: 'SingleQuestion',
	3: 'MultyOptionsQuestion',
	4: 'NumericQuestion',
	5: 'DateTimeQuestion',
	6: 'GpsCoordinateQuestion',
	7: 'TextQuestion',
	9: 'TextListQuestion',
	10: 'QRBarcodeQuestion',
	11: 'MultimediaQuestion'
}

VARIABLE_TYPES = {
	3: 'Boolean'
}

ROSTER_TYPES = {
	1: 'Fixed set of items'
}


CLASSTYPE_MACRO = "System.Collections.Generic.Dictionary`2[[System.Guid, mscorlib],[WB.Core.SharedKernels.SurveySolutions.Documents.Macro, WB.Core.SharedKernels.Questionnaire]], mscorlib"
CLASSTYPE_LOOKUPTABLE = "System.Collections.Generic.Dictionary`2[[System.Guid, mscorlib],[WB.Core.SharedKernels.SurveySolutions.Documents.LookupTable, WB.Core.SharedKernels.Questionnaire]], mscorlib"


from .base import HQBase
from .exceptions import NotFoundError, GraphQLError
from .headquarters_schema import headquarters_schema, InterviewFilter, Interview
from .utils import fix_qid
from sgqlc.operation import Operation
from typing import Generator

class InterviewsApi(HQBase):
    _apiprefix = "/api/v1/interviews"

    @fix_qid(expects='hex', param_name='questionnaire_id')
    def get_list(self, fields: list = (), **kwargs) -> Generator[Interview, None, None]:
        where = InterviewFilter(**kwargs)
        take = 20
        skip = 0
        filtered_count = 21
        if not fields:
            fields = (
                'id',
                'questionnaire_id',
                'questionnaire_version',
                'assignment_id',
                'responsible_id',
                'errors_count',
                'status',
            )
        while skip < filtered_count:  
            op = Operation(headquarters_schema.HeadquartersQuery)  
            q = op.interviews(take=take, skip=skip, where=where)
            q.__fields__('filtered_count')
            q.nodes.__fields__(*fields)
            print(op)
            cont = self.endpoint(op)
            errors = cont.get('errors')
            if errors:
                raise GraphQLError(errors[0]['message'])
            res = (op + cont).interviews

            filtered_count = res.filtered_count
            yield from res.nodes
            skip += take

    def get_info(self, interviewid):
        path = self.url + '/{}'.format(interviewid)
        return self._make_call('get', path)

    def delete(self, interviewid, comment=''):
        return self._change_status(action='delete', interviewid=interviewid, comment=comment)

    def answers(self, interviewid):
        pass

    def approve(self, interviewid, comment=''):
        return self._change_status(action='approve', interviewid=interviewid, comment=comment)

    def assign(self, interviewid, responsibleid, responsiblename=''):
        return self._reassign(action='assign', interviewid=interviewid, responsibleid=responsibleid, responsiblename=responsiblename)

    def assign_supervisor(self, interviewid, responsibleid, responsiblename=''):
        return self._reassign(action='assignsupervisor', interviewid=interviewid, responsibleid=responsibleid, responsiblename=responsiblename)

    def comment(self, interviewid, comment, questionid=None, variable=''):
        pass

    def history(self, interviewid):
        path = self.url + '/{}'.format(interviewid) + '/history'
        return self._make_call('get', path)

    def hqapprove(self, interviewid, comment=''):
        return self._change_status(action='hqapprove', interviewid=interviewid, comment=comment)

    def hqreject(self, interviewid, comment=''):
        return self._change_status(action='hqreject', interviewid=interviewid, comment=comment)

    def hqunapprove(self, interviewid, comment=''):
        return self._change_status(action='hqunapprove', interviewid=interviewid, comment=comment)

    def reject(self, interviewid, comment=''):
        return self._change_status(action='reject', interviewid=interviewid, comment=comment)

    def stats(self, interviewid):
        pass

    def _change_status(self, interviewid, action, comment=''):
        path = self.url + '/{}/{}?comment={}'.format(interviewid, action, comment)
        return self._make_call('patch', path)

    def _reassign(self, interviewid, action, responsibleid, responsiblename=''):
        path = self.url + '/{}/{}'.format(interviewid, action)
        payload = {
            'ResponsibleId': responsibleid,
            'ResponsibleName': responsiblename
        }
        r = self._make_call('patch', path, json = payload)
        return True


from .headquarters import Client
from .assignments import AssignmentsApi
from .export import ExportApi
from .interviews import InterviewsApi
from .questionnaires import QuestionnairesApi
from .settings import SettingsApi
from .users import UsersApi
from . import designer
from .__about__ import __version__